/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: scout_thread_arch.h,v $
 *
 * HISTORY
 * $Log: scout_thread_arch.h,v $
 * Revision 1.1  1996/01/30  22:43:17  slm
 * Initial revision
 *
 * Revision 1.3  1994/11/20  01:31:34  davidm
 * Finalized FPU support.
 *
 * Revision 1.2  1994/06/11  00:02:55  menze
 * Added copyright notice
 *
 * Revision 1.1  1994/05/15  00:03:03  davidm
 * Initial revision
 */
#ifndef scout_thread_arch_h
#define scout_thread_arch_h

/*
 * Alpha register usage (OSF/1 1.3):
 *
 * Callee saved registers:	r9-r14, r26, f2-f9
 *
 *	r0  = function return value (non-fp)
 *	r15 = frame pointer (if has not been optimized away)
 *	r29 = global pointer
 *	r26 = return PC
 *	r30 = stack pointer (SP)
 *	r31 = always 0
 *	f0  = function return value (fp)
 *	f31 = always 0.0
 */

#define MAGIC_COOKIE		0xFEEDF00DBEC0FFEE
#define MIN_STACK_SIZE		(1024)	/* enough for a single printf() */
#define MIN_USEFUL_STACK_SIZE	(2*MIN_STACK_SIZE)
#define DEFAULT_STACK_SIZE	(8*MIN_USEFUL_STACK_SIZE)
#define FPU_STATE_SIZE		(9*sizeof(long))

/*
 * Switch to stack TO_SP, discarding the old stack pointer and invoke
 * start().  On the Alpha, a return PC of 0 indicates a top-level
 * function.  So unlike in a normal procedure call, we set RA ($26) to 0:
 */
#define SWITCH_STACK_AND_START(to_sp,start) \
{ \
    asm volatile ("mov %0, $sp
		   mov $31, $26
		   br $31, start..ng" : : "r"(to_sp), "i"(start)); \
}

/*
 * Save necessary state on stack, save new stack pointer to
 * variable SAVED_SP:
 */
#define SAVE_STATE(saved_sp) \
{ \
    asm volatile \
      ("lda $sp, -64($sp)	# allocate space for state on stack
	stq  $9,  0($sp)	# save callee-saved registers
	stq $10,  8($sp)	# r9-r15 and r26
	stq $11, 16($sp)
	stq $12, 24($sp)
	stq $13, 32($sp)
	stq $14, 40($sp)
	stq $fp, 48($sp)	# aka r15
	stq $26, 56($sp)	# save return PC
	stq $sp, %0		# save sp to saved_sp
        " : "m=" (saved_sp) : : "memory"); \
} /* SAVE_STATE */


/*
 * Load state stored at FROM_SP.
 */
#define LOAD_STATE(from_sp) \
{ \
    asm volatile \
      ("ldq $sp, %0 		# restore sp from from_sp
	ldq  $9,  0($sp)	# restore callee-saved registers
	ldq $10,  8($sp)	# r9-r15 and r26
	ldq $11, 16($sp)
	ldq $12, 24($sp)
	ldq $13, 32($sp)
	ldq $14, 40($sp)
	ldq $fp, 48($sp)	# aka r15
	ldq $26, 56($sp)	# save return PC
	lda $sp, +64($sp)	# pop saved state
	" : : "m" (from_sp)); \
} /* LOAD_STATE */

#define SAVE_FPU_STATE(fpu_state) \
{ \
    asm volatile \
      ("trapb			# take outstanding traps now
	stt $f2, 0(%0)
	mf_fpcr $f2, $f2, $f2
	stt $f3, 8(%0)
	stt $f4, 16(%0)
	stt $f5, 24(%0)
	stt $f6, 32(%0)
	stt $f7, 40(%0)
	stt $f8, 48(%0)
	stt $f9, 56(%0)
	stt $f2, 64(%0)		# save fpcr
	" : : "r" (fpu_state) : "memory"); \
}
       
#define LOAD_FPU_STATE(fpu_state) \
{ \
    asm volatile \
      ("ldt $f9, 64(%0)		# load fpcr
	ldt $f2,  0(%0)
	ldt $f3,  8(%0)
	ldt $f4, 16(%0)
	ldt $f5, 24(%0)
	ldt $f6, 32(%0)
	ldt $f7, 40(%0)
	ldt $f8, 48(%0)
	mt_fpcr $f9, $f9, $f9	# install fpcr
	ldt $f9, 56(%0)
	trapb			# ensure fpcr is updated before traps are taken
	" : : "r" (fpu_state) : "memory"); \
}

#define FPU_ON \
{ \
   asm volatile ("mov 1, $16
		  call_pal %0" \
		 :: "i"(PAL_wrfen) \
		 : "memory", "1", "16", "22", "23", "24", "25"); \
}

#define FPU_OFF \
{ \
   asm volatile ("mov $31, $16
		  call_pal %0" \
		 :: "i"(PAL_wrfen) \
		 : "memory", "1", "16", "22", "23", "24", "25"); \
}

/* this is for debugging purposes only: */
#define GET_SAVED_PC(sp)	((void*)(((u_long*)(sp))[7]))

#endif /* scout_thread_arch_h */
