/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: alpha_init.c,v $
 *
 * HISTORY
 * $Log: alpha_init.c,v $
 * Revision 1.9  1996/02/02  15:54:45  slm
 * Updated copyright and version.
 *
 * Revision 1.8  1995/03/10  19:31:51  davidm
 * EntSys handler added.
 *
 * Revision 1.7  1994/11/20  06:43:20  davidm
 * Added fpdata/__dpml_capture_context() hack to enable use of libm.a.
 *
 * Revision 1.6  1994/11/20  01:36:05  davidm
 * Added FPU support and various minor fixes.
 *
 * Revision 1.5  1994/10/07  00:46:38  davidm
 * (set_branch_history_table): new function to enable/disable branch-history
 * 	table.
 *
 * (main_): fixed for OSF/1 v3.0; clearing the BSS broke because edata
 * 	points to the beginning of the *data* segment rather than the
 * 	beginning of the *bss* segment---smells like a serious bug.
 * 	Fortunately, _fbss does correctly point to the beginning of
 * 	the bss.
 *
 * Revision 1.4  1994/06/11  00:02:37  menze
 * Added copyright notice
 *
 * Revision 1.3  1994/05/27  22:40:06  druschel
 * Moved argument query to init(), where threads are initialized already.
 *
 * Revision 1.2  1994/05/14  23:38:15  davidm
 * More cleanup.
 *
 * Revision 1.1  1994/04/20  00:53:51  davidm
 * Initial revision
 */
/*
 * Alpha initialization.  Function "main_" is the first C function
 * to get invoked.
 *
 * This is based on information published in:
 *
 * VMS for Alpha Platforms: Internals and Data Structures, Volume 1,
 * Chapter 33 (Bootstrap Processing), DEC Order #: EY-L466E-P1.
 */
#include <ctype.h>
#include <sys/types.h>
#include <machine/rpb.h>
#include <machine/prom.h>
#include "scout_boot.h"
#include "scout_config.h"
#include "scout_device.h"
#include "scout_malloc.h"
#include "scout_thread.h"
#include "scout_time.h"
#include "scout_util.h"
#include "alpha_pal.h"
#include "alpha_prom.h"
#include "alpha_misc.h"
#include "alpha_console.h"
#include "kgdb.h"

#define MAX(a,b)	((a) > (b) ? (a) : (b))
#define MIN(a,b)	((a) < (b) ? (a) : (b))

struct rpb		*rpb;
struct rpb_percpu	*percpu;
struct rpb_ctb		*ctb;
struct rpb_crb		*crb;
struct rpb_mdt		*mdt;
struct rpb_cluster	*cluster;
struct rpb_map		*map;
long			console_unit = 0;
long			scout_page_size;
u_int			scout_boot_flags;

extern char		*testArgs;


void
init(void *arg)
{
    static char cmdline[256] = "";

    flamingo_init();

    /* enable interrupts: */
    scout_swap_ipl(0);

    if (scout_boot_flags & SCOUT_BOOT_F_QUERY) {
	printf("scout: ");
	gets(cmdline);
	if (cmdline[0]) {
	    testArgs = cmdline;
	} /* if */
    } /* if */

    xkInit();
} /* init */


static void
set_handlers(void)
{
    extern char _gp[];
    extern int instruction_fault[], mm_fault[], unaligned_fault[];
    extern int arith_fault[], syscall[];
#ifdef SCOUT_ROLLFORWARD
    extern int interrupt_maybe[];
#else
    extern int interrupt[];
#endif

    /* setup interrupt/exception handlers: */
    wrkgp((long) _gp);
    wrent(PAL_EntIF,  (long) instruction_fault);
    wrent(PAL_EntMM,  (long) mm_fault);
    wrent(PAL_EntUna, (long) unaligned_fault);
#ifdef SCOUT_ROLLFORWARD
    wrent(PAL_EntInt, (long) interrupt_maybe);
#else
    wrent(PAL_EntInt, (long) interrupt);
#endif
    wrent(PAL_EntArith, (long) arith_fault);
    wrent(PAL_EntSys, (long) syscall);
} /* set_handlers */


static void
set_branch_history_table(int on)
{
    u_long iccsr;

    if (on) {
	asm volatile ("
	    .long 0x64210082	# hw_mfpr/p t0, $iccsr
	    .long 0x205f0040	# lda t1, 0x40(zero)
	    .long 0x48441722	# sll t1, 32, t1
	    .long 0x44220401	# bis t0, t1, t0
	    .long 0x742100a2	# hw_mtpr/pi t0, $iccsr
	    mov $1, %0
	    " : "r="(iccsr) :: "1", "2", "3");
    } else {
	asm volatile ("
	    .long 0x64210082	# hw_mfpr/p t0, $iccsr
	    .long 0x205f0040	# lda t1, 0x40(zero)
	    .long 0x48441722	# sll t1, 32, t1
	    .long 0x44220101	# bic t0, t1, t0
	    .long 0x742100a2	# hw_mtpr/pi t0, $iccsr
	    mov $1, %0
	    " : "r="(iccsr) :: "1", "2", "3");
    } /* if */
    printf("iccsr=#%lx\n", iccsr);
} /* set_branch_history_table */


long
alpha_set_icache_streaming(int on)
{
    register u_long actl asm("3");

    if (on) {
	actl = 0x43e;
	asm volatile ("
	    .long 0x7463004e	# hw_mtpr/a t2, $abox_ctl
	    " :: "r"(actl));
    } else {
	actl = 0x436;
	asm volatile ("
	    .long 0x7463004e	# hw_mtpr/a t2, $abox_ctl
	    " :: "r"(actl));
    } /* if */
    return actl;
} /* alpha_set_icache_streaming */


#ifdef SCOUT_FPU_SUPPORT
/*
 * Scout does not currently support C runtime exception handling as
 * expected by libm.a.  However, with the following hack, using libm
 * functions works fine as long as no exceptions are generated.
 * sqrt(-3), 1.0/0.0, or exp(10e5), are all examples of exception
 * generating computations.
 */

int fpdata;

void
__dpml_capture_context()
{
    scout_panic("__dpml_capture_context: ouch---we got an FP exception!");
} /* __dpml_capture_context */

#endif /* SCOUT_FPU_SUPPORT */


void
main_(int argc, char **argv)
{
    int i;
    long retval;
    PROM_Result res;
    PALcode_Revision vmsrev, osfrev;
    static char buf[MAX_ENVIRON_LENGTH];
    char *start_addr, *end_addr, *left_edge, *right_edge;
    extern char _end[], _fbss[];

    /* clear BSS: */
    bzero(_fbss, _end - _fbss);

#ifdef SCOUT_FPU_SUPPORT
#   define PAL_mtpr_fen	0x0c
    /*
     * GCC uses fp regs as temporary storage, so better make sure
     * those won't cause an FPU fault.  Because we're still in VMS
     * land here, we have to use the mtpr_fen pal-call instead of the
     * PAL_wrfen.  The FPU will be disabled with the first context
     * switch to a real thread, after that things roll they way they
     * should.
     */
    asm volatile("mov 1, $16; call_pal %0" :: "i"(PAL_mtpr_fen) :
		 "0", "1", "16", "17");
#endif

    /*
     * Setup some pointers for convenient access to the hardware
     * restart parameter block:
     */
    rpb = (struct rpb*) HWRPB_ADDR;
    percpu = (struct rpb_percpu*) ((char*) rpb + rpb->rpb_percpu_off);
    ctb = (struct rpb_ctb*) ((char*) rpb + rpb->rpb_ctb_off);
    crb = (struct rpb_crb*) ((char*) rpb + rpb->rpb_crb_off);
    mdt = (struct rpb_mdt*) ((char*) rpb + rpb->rpb_mdt_off);
    map = &crb->rpb_map[0];
    cluster = &mdt->rpb_cluster[0];

    /* get console unit so we can print error messages: */
    res.bits = PROM_GETENV(PROM_E_TTY_DEV, buf, sizeof(buf));
    buf[res.u.retval] = '\0';

    console_unit = buf[0] - '0';
    printf("console_unit=%s\n", buf);

    /*
     * Switch from VMS to OSF/1 PALcode:
     */
    vmsrev.bits = percpu[rpb->rpb_cpuid].rpb_palrev;
    osfrev.bits = percpu[rpb->rpb_cpuid].rpb_palrev_avail[PALvar_OSF1];
    printf("switching PALcode from VMS version %d.%d to OSF/1 version %d.%d\n",
	   vmsrev.u.vers, vmsrev.u.rev, osfrev.u.vers, osfrev.u.rev);
    retval = prom_switch_to_osf_pal_code();
    if (retval != 0) {
	scout_abort("prom_switch_to_osf_pal_code() returned %d\n", retval);
    } /* if */

    set_handlers();

    /* these require the HWE bit to be on: */
    set_branch_history_table(1);
    printf("abox control=%#lx\n", alpha_set_icache_streaming(1));

    /* get bootflags: */
    res.bits = PROM_GETENV(PROM_E_BOOTED_OSFLAGS, buf, sizeof(buf));
    for (i = 0; i < res.u.retval; ++i) {
	switch (tolower(buf[i])) {
	  case 'a': scout_boot_flags |= SCOUT_BOOT_F_AUTO; break;
	  case 'k': scout_boot_flags |= SCOUT_BOOT_F_DEBUG; break;
	  case 'd': scout_boot_flags |= SCOUT_BOOT_F_DUMP; break;
	  case 'i': scout_boot_flags |= SCOUT_BOOT_F_QUERY; break;
	  case 'v': scout_boot_flags |= SCOUT_BOOT_F_VERBOSE; break;
	  case '0': case ',': break;
	  default:
	    printf("Unknown boot-option `%c'\n", buf[i]);
	    break;
	} /* switch */
    } /* for */

    /* initialize tracing variables: */
    {
	extern int trace_turbo_channel;
	extern int trace_scout_device;
	extern int trace_scout_malloc;
	extern int trace_scout_thread;

	trace_turbo_channel = 0;
	trace_scout_device  = 0;
	trace_scout_malloc  = 0;
	trace_scout_thread  = 0;
    }

    scout_ticks_per_second = rpb->rpb_clock / 4096;
    scout_page_size = rpb->rpb_pagesize;

    printf("interval clock freq=%ld Hz, cycle counter clock freq=%ld Hz\n",
	   scout_ticks_per_second, rpb->rpb_counter);

    if (scout_boot_flags & SCOUT_BOOT_F_VERBOSE) {
	printf("play leds: "
#ifdef SCOUT_PLAY_LEDS
	       "on"
#else
	       "off"
#endif
	       "; kgdb: "
#ifdef SCOUT_KGDB
	       "on"
#else
	       "off"
#endif
	       "; rollforward: "
#ifdef SCOUT_ROLLFORWARD
	       "on"
#else
	       "off"
#endif
	       "; predictions: "
#ifdef __PREDICTIONS__
	       "on"
#else
	       "off"
#endif
	       "; profiling: "
#ifdef SCOUT_PROFILE
	       "on"
#else
	       "off"
#endif
	       "\n");
    } /* if */

#ifdef SCOUT_KGDB
    /* initialize kgdb: */
    kgdb_set_debug_traps();
    if (scout_boot_flags & SCOUT_BOOT_F_DEBUG) {
	printf("kgdb: waiting for remote gdb to connect...\n");
	kgdb_breakpoint();
    } /* if */
#endif /* SCOUT_KGDB */

    /* tell memory allocator about available physical memory: */
    for (i = 0; i < mdt->rpb_numcl; ++i) {
	if (cluster[i].rpb_usage == CLUSTER_USAGE_OS) {
	    start_addr = (char*)
	      P_TO_K0SEG(cluster[i].rpb_pfn * scout_page_size);
	    end_addr = start_addr + scout_page_size * cluster[i].rpb_pfncount;
	    /* find intersection between kernel and this cluster: */
	    left_edge  = MAX(start_addr, (char*)FIRST_TEXT);
	    right_edge = MIN(end_addr, _end);
	    if (left_edge < right_edge) {
		/* intersection is non-empty: */
		if (start_addr < (char*)FIRST_TEXT) {
		    scout_malloc_expand(start_addr,
					(char*)FIRST_TEXT - start_addr);
		} /* if */
		if (end_addr > _end) {
		    scout_malloc_expand(_end, end_addr - _end);
		} /* if */
	    } else {
		scout_malloc_expand(start_addr, end_addr - start_addr);
	    } /* if */
	} /* if */
    } /* for */

    scout_thread_init(init, 0);
    scout_panic(__FILE__": initializer thread returned!");
} /* main_ */

			/*** end of alpha_startup.c ***/
